from csle_common.dao.docker.docker_container_metadata import DockerContainerMetadata
from csle_common.dao.docker.docker_env_metadata import DockerEnvMetadata


class TestDockerDaoSuite:
    """
    Test suite for Docker data access objects (DAOs)
    """

    def test_docker_container_metadata(self, example_docker_container_metadata: DockerContainerMetadata) -> None:
        """
        Tests creation and dict conversion of the DockerContainerMetadata DAO

        :param example_docker_container_metadata: an example DockerContainerMetadata
        :return: None
        """
        assert isinstance(example_docker_container_metadata.to_dict(), dict)
        assert isinstance(DockerContainerMetadata.from_dict(example_docker_container_metadata.to_dict()),
                          DockerContainerMetadata)
        assert DockerContainerMetadata.from_dict(example_docker_container_metadata.to_dict()).to_dict() == \
               example_docker_container_metadata.to_dict()
        assert DockerContainerMetadata.from_dict(example_docker_container_metadata.to_dict()) == \
               example_docker_container_metadata

    def test_docker_env_metadata(self, example_docker_env_metadata: DockerEnvMetadata) -> None:
        """
        Tests creation and dict conversion of the DockerEnvMetadata DAO

        :param example_docker_env_metadata: an example DockerEnvMetadata
        :return: None
        """
        assert isinstance(example_docker_env_metadata.to_dict(), dict)
        assert isinstance(DockerEnvMetadata.from_dict(example_docker_env_metadata.to_dict()), DockerEnvMetadata)
        assert DockerEnvMetadata.from_dict(example_docker_env_metadata.to_dict()).to_dict() \
               == example_docker_env_metadata.to_dict()
        assert DockerEnvMetadata.from_dict(example_docker_env_metadata.to_dict()) == example_docker_env_metadata



"""
Type of packet losses on a container's network interface
"""
from enum import IntEnum


class PacketLossType(IntEnum):
    """
    Enum representing the different types of packet losses to emulate on a container
    """
    RANDOM = 0
    STATE = 1
    GEMODEL = 2



"""
Network configuration of a container in the emulation
"""
from typing import Dict, Any
from csle_common.dao.emulation_config.packet_loss_type import PacketLossType
from csle_common.dao.emulation_config.packet_delay_distribution_type import PacketDelayDistributionType
import csle_common.constants.constants as constants
from csle_base.json_serializable import JSONSerializable


class NodeNetworkConfig(JSONSerializable):
    """
    A DTO object representing the network configuration of a specific container in an emulation environment
    """

    def __init__(self, interface: str = constants.NETWORKING.ETH0,
                 limit_packets_queue: int = 30000, packet_delay_ms: float = 0.1,
                 packet_delay_jitter_ms: float = 0.025,
                 packet_delay_correlation_percentage: float = 25,
                 packet_delay_distribution: PacketDelayDistributionType = PacketDelayDistributionType.PARETO,
                 packet_loss_type: PacketLossType = PacketLossType.GEMODEL,
                 packet_loss_rate_random_percentage: float = 2,
                 packet_loss_random_correlation_percentage: float = 25,
                 loss_state_markov_chain_p13: float = 0.1, loss_state_markov_chain_p31: float = 0.1,
                 loss_state_markov_chain_p32: float = 0.1, loss_state_markov_chain_p23: float = 0.1,
                 loss_state_markov_chain_p14: float = 0.1, loss_gemodel_p: float = 0.0001,
                 loss_gemodel_r: float = 0.999,
                 loss_gemodel_h: float = 0.0001, loss_gemodel_k: float = 0.9999,
                 packet_corrupt_percentage: float = 0.00001, packet_corrupt_correlation_percentage: float = 25,
                 packet_duplicate_percentage: float = 0.00001, packet_duplicate_correlation_percentage: float = 25,
                 packet_reorder_percentage: float = 0.0025, packet_reorder_correlation_percentage: float = 25,
                 packet_reorder_gap: int = 5,
                 rate_limit_mbit: float = 100,
                 packet_overhead_bytes: int = 0, cell_overhead_bytes: int = 0):
        """
        Initializes the DTO

        :param interface: the name of the network interface
        :param limit_packets_queue: Maximum number of packets that the output queue (queueing discpline) stores
                                    Note that to be able to rate limit at relatively high speeds, this queue must be
                                    quite large (FIFO by default)
        :param packet_delay_ms: a delay that is added to outgoing packets (specified in ms)
        :param packet_delay_jitter_ms: the jitter of the dealy added to outgoing packets (specified in ms)
        :param packet_delay_correlation_percentage: the amount of correlation in delay of outgoing packets
                                                   (specified in %)
                                                   for example delay 100ms jitter=10ms correlation=25% causes delay
                                                   to be 100ms +- 10ms with successive packets depending 25%
                                                   on each other
        :param packet_delay_distribution: the delay distribution, defaults to Gaussian, but can also be uniform,
                                                normal, pareto, or paretonormal
        :param packet_loss_type: the type of packet loss (random, state, or gemodel)
        :param packet_loss_rate_random_percentage: The packet loss percent when using the random packet loss type
        :param packet_loss_random_correlation_percentage: The packet loss correlation when using the random
                                                          packet loss type
        :param loss_state_markov_chain_p13: The 1->3 transition probability when using the 2-state Markov
                                            chain state loss type
        :param loss_state_markov_chain_p31: The 3->1 transition probability when using the 2,3,or 4-state
                                            Markov chain state loss type
        :param loss_state_markov_chain_p32: The 3->2 transition probability when using the 3,or 4-state Markov chain
                                            state loss type
        :param loss_state_markov_chain_p23: The 2->3 transition probability when using the 3,or 4-state Markov chain
                                            state loss type
        :param loss_state_markov_chain_p14: The 1->4 transition probability when using the 4-state Markov chain state
                                            loss type
        :param loss_gemodel_p: The p parameter when using the Bernoulli,Simple Gilbert or Gilbert model
                               in the gemodel loss type
        :param loss_gemodel_r: The p parameter when using the Simple Gilbert or Gilbert-Elliot model in the
                               gemodel loss type
        :param loss_gemodel_h: The h parameter when using the Simple Gilbert or Gilbert-Elliot model
                               in the gemodel loss type
        :param loss_gemodel_k: The k parameter when using the Gilbert-Elliot model in the gemodel loss type
        :param packet_corrupt_percentage: The percentage of corrupt outgoing packets
        :param packet_corrupt_correlation_percentage: The correlation of corrupt outgoing packets
        :param packet_duplicate_percentage: The percentage of duplicated outgoing packets
        :param packet_duplicate_correlation_percentage: The correlation of duplicated outgoing packets
        :param packet_reorder_percentage: The percentage of reordered packets
        :param packet_reorder_correlation_percentage: The correlation of reordered packets
        :param packet_reorder_gap: The packet reordering gap
        :param rate_limit_mbit: The bandwidth limit
        :param packet_overhead_bytes: The emulated packet overhead
        :param cell_overhead_bytes: The emulated cell overhead
        """
        self.interface = interface
        self.limit_packets_queue = limit_packets_queue
        self.packet_delay_ms = packet_delay_ms
        self.packet_delay_jitter_ms = packet_delay_jitter_ms
        self.packet_delay_correlation_percentage = packet_delay_correlation_percentage
        self.packet_delay_distribution = packet_delay_distribution
        self.packet_loss_type = packet_loss_type
        self.packet_loss_rate_random_percentage = packet_loss_rate_random_percentage
        self.packet_loss_random_correlation_percentage = packet_loss_random_correlation_percentage
        self.loss_state_markov_chain_p13 = loss_state_markov_chain_p13
        self.loss_state_markov_chain_p31 = loss_state_markov_chain_p31
        self.loss_state_markov_chain_p32 = loss_state_markov_chain_p32
        self.loss_state_markov_chain_p23 = loss_state_markov_chain_p23
        self.loss_state_markov_chain_p14 = loss_state_markov_chain_p14
        self.loss_gemodel_p = loss_gemodel_p
        self.loss_gemodel_r = loss_gemodel_r
        self.loss_gemodel_h = loss_gemodel_h
        self.loss_gemodel_k = loss_gemodel_k
        self.packet_corrupt_percentage = packet_corrupt_percentage
        self.packet_corrupt_correlation_percentage = packet_corrupt_correlation_percentage
        self.packet_duplicate_percentage = packet_duplicate_percentage
        self.packet_duplicate_correlation_percentage = packet_duplicate_correlation_percentage
        self.packet_reorder_percentage = packet_reorder_percentage
        self.packet_reorder_correlation_percentage = packet_reorder_correlation_percentage
        self.packet_reorder_gap = packet_reorder_gap
        self.rate_limit_mbit = rate_limit_mbit
        self.packet_overhead_bytes = packet_overhead_bytes
        self.cell_overhead_bytes = cell_overhead_bytes

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeNetworkConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        obj = NodeNetworkConfig(
            interface=d["interface"],
            limit_packets_queue=d["limit_packets_queue"],
            packet_delay_ms=d["packet_delay_ms"],
            packet_delay_jitter_ms=d["packet_delay_jitter_ms"],
            packet_delay_correlation_percentage=d["packet_delay_correlation_percentage"],
            packet_delay_distribution=d["packet_delay_distribution"],
            packet_loss_type=d["packet_loss_type"],
            packet_loss_rate_random_percentage=d["packet_loss_rate_random_percentage"],
            packet_loss_random_correlation_percentage=d["packet_loss_random_correlation_percentage"],
            loss_state_markov_chain_p13=d["loss_state_markov_chain_p13"],
            loss_state_markov_chain_p31=d["loss_state_markov_chain_p31"],
            loss_state_markov_chain_p32=d["loss_state_markov_chain_p32"],
            loss_state_markov_chain_p23=d["loss_state_markov_chain_p23"],
            loss_state_markov_chain_p14=d["loss_state_markov_chain_p14"],
            loss_gemodel_p=d["loss_gemodel_p"],
            loss_gemodel_r=d["loss_gemodel_r"],
            loss_gemodel_h=d["loss_gemodel_h"],
            loss_gemodel_k=d["loss_gemodel_k"],
            packet_corrupt_percentage=d["packet_corrupt_percentage"],
            packet_corrupt_correlation_percentage=d["packet_corrupt_correlation_percentage"],
            packet_duplicate_percentage=d["packet_duplicate_percentage"],
            packet_duplicate_correlation_percentage=d["packet_duplicate_correlation_percentage"],
            packet_reorder_percentage=d["packet_reorder_percentage"],
            packet_reorder_correlation_percentage=d["packet_reorder_correlation_percentage"],
            packet_reorder_gap=d["packet_reorder_gap"],
            rate_limit_mbit=d["rate_limit_mbit"],
            packet_overhead_bytes=d["packet_overhead_bytes"],
            cell_overhead_bytes=d["cell_overhead_bytes"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["interface"] = self.interface
        d["limit_packets_queue"] = self.limit_packets_queue
        d["packet_delay_ms"] = self.packet_delay_ms
        d["packet_delay_jitter_ms"] = self.packet_delay_jitter_ms
        d["packet_delay_correlation_percentage"] = self.packet_delay_correlation_percentage
        d["packet_delay_distribution"] = self.packet_delay_distribution
        d["packet_loss_type"] = self.packet_loss_type
        d["packet_loss_rate_random_percentage"] = self.packet_loss_rate_random_percentage
        d["packet_loss_random_correlation_percentage"] = self.packet_loss_random_correlation_percentage
        d["loss_state_markov_chain_p13"] = self.loss_state_markov_chain_p13
        d["loss_state_markov_chain_p31"] = self.loss_state_markov_chain_p31
        d["loss_state_markov_chain_p32"] = self.loss_state_markov_chain_p32
        d["loss_state_markov_chain_p23"] = self.loss_state_markov_chain_p23
        d["loss_state_markov_chain_p14"] = self.loss_state_markov_chain_p14
        d["loss_gemodel_p"] = self.loss_gemodel_p
        d["loss_gemodel_r"] = self.loss_gemodel_r
        d["loss_gemodel_h"] = self.loss_gemodel_h
        d["loss_gemodel_k"] = self.loss_gemodel_k
        d["packet_corrupt_percentage"] = self.packet_corrupt_percentage
        d["packet_corrupt_correlation_percentage"] = self.packet_corrupt_correlation_percentage
        d["packet_duplicate_percentage"] = self.packet_duplicate_percentage
        d["packet_duplicate_correlation_percentage"] = self.packet_duplicate_correlation_percentage
        d["packet_reorder_percentage"] = self.packet_reorder_percentage
        d["packet_reorder_correlation_percentage"] = self.packet_reorder_correlation_percentage
        d["packet_reorder_gap"] = self.packet_reorder_gap
        d["rate_limit_mbit"] = self.rate_limit_mbit
        d["packet_overhead_bytes"] = self.packet_overhead_bytes
        d["cell_overhead_bytes"] = self.cell_overhead_bytes
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"limits_packets_queue:{self.limit_packets_queue}, packet_delay_ms:{self.packet_delay_ms}, " \
               f"packet_delay_jitter_ms:{self.packet_delay_jitter_ms}, " \
               f"packet_delay_correlation_percentage: {self.packet_delay_correlation_percentage}, " \
               f"packet_delay_distribution: {self.packet_delay_distribution}, " \
               f"packet_loss_type:{self.packet_loss_type}, " \
               f"packet_loss_rate_random_percentage: {self.packet_loss_rate_random_percentage}, " \
               f"packet_loss_correlation_percentage: {self.packet_loss_random_correlation_percentage}, " \
               f"loss_state_markov_chain_p13: {self.loss_state_markov_chain_p13}, " \
               f"loss_state_markov_chain_p31: {self.loss_state_markov_chain_p31}, " \
               f"loss_state_markov_chain_p32: {self.loss_state_markov_chain_p32}, " \
               f"loss_state_markov_chain_p23: {self.loss_state_markov_chain_p23}, " \
               f"loss_state_markov_chain_p14: {self.loss_state_markov_chain_p14}, " \
               f"loss_gemodel_p:{self.loss_gemodel_p}, loss_gemodel_r: {self.loss_gemodel_r}, " \
               f"loss_gemodel_h:{self.loss_gemodel_h}, loss_gemodel_k:{self.loss_gemodel_k}, " \
               f"packet_corrupt_percentage:{self.packet_corrupt_percentage}, " \
               f"packet_corrupt_correlation_percentage:{self.packet_corrupt_correlation_percentage}, " \
               f"packet_duplicate_percentage:{self.packet_duplicate_percentage}, " \
               f"packet_duplicate_correlation_percentage:{self.packet_duplicate_correlation_percentage}, " \
               f"packet_reorder_percentage:{self.packet_reorder_percentage}, " \
               f"packet_reorder_correlation_percentage:{self.packet_reorder_correlation_percentage}, " \
               f"packet_reorder_gap: {self.packet_reorder_gap}" \
               f"rate_limit_mbit:{self.rate_limit_mbit}, packet_overhead_bytes:{self.packet_overhead_bytes}," \
               f"cell_overhead_bytes:{self.cell_overhead_bytes}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeNetworkConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeNetworkConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeNetworkConfig":
        """
        :return: a copy of the DTO
        """
        return NodeNetworkConfig.from_dict(self.to_dict())

    @staticmethod
    def schema() -> "NodeNetworkConfig":
        """
        :return: get the schema of the DTO
        """
        return NodeNetworkConfig()



from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
import pytest
from csle_common.controllers.snort_ids_controller import SnortIDSController


class TestSnortIdsControllerSuite:
    """
    Test snort_ids_controller
    """

    @pytest.fixture(autouse=True)
    def emulation_env_config_setup(self) -> None:
        """
        Set up emulation environment configuration

        :return: None
        """
        self.logger = MagicMock()
        constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES = ["container1", "container2"]
        container1 = MagicMock()
        container1.name = "container1"
        container1.physical_host_ip = "192.168.1.1"
        container1.docker_gw_bridge_ip = "10.0.0.1"
        container1.get_ips.return_value = ["10.0.1.1"]
        container1.get_full_name.return_value = "container-1"
        container2 = MagicMock()
        container2.name = "container2"
        container2.physical_host_ip = "192.168.1.2"
        container2.docker_gw_bridge_ip = "10.0.0.2"
        container2.get_ips.return_value = ["10.0.1.2"]
        container2.get_full_name.return_value = "container-2"
        containers = [container1, container2]
        emulation_env_config = MagicMock()
        emulation_env_config.containers_config.containers = containers
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_port = 50051
        emulation_env_config.execution_id = "12345"
        emulation_env_config.level = "1"
        self.emulation_env_config = emulation_env_config

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_ids")
    def test_start_snort_idses(self, mock_start_snort_ids) -> None:
        """
        Test utility function for starting the Snort IDSes

        :param mock_start_snort_ids: mock_start_snort_ids
        :return: None
        """
        SnortIDSController.start_snort_idses(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        self.logger.info.assert_called_once_with(
            f"Starting the Snort IDS on IP: "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip} ("
            f"{self.emulation_env_config.containers_config.containers[0].get_ips()[0]}, "
            f"{self.emulation_env_config.containers_config.containers[0].get_full_name()})")
        mock_start_snort_ids.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.stop_snort_ids")
    def test_stop_snort_idses(self, mock_stop_snort_ids) -> None:
        """
        Test utility function for stopping the Snort IDSes

        :param mock_stop_snort_ids: mock_stop_snort_ids
        :return: None
        """
        SnortIDSController.stop_snort_idses(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_stop_snort_ids.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_manager")
    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController."
           "get_snort_idses_monitor_threads_statuses_by_ip_and_port")
    @patch("csle_collector.snort_ids_manager.query_snort_ids_manager.start_snort_ids")
    @patch("grpc.insecure_channel")
    def test_start_snort_ids(self, mock_insecure_channel, mock_start_snort_ids, mock_get_statuses, mock_start_manager) \
            -> None:
        """
        Test utility function for starting the Snort IDS on a specific IP

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_snort_ids: mock_start_snort_ids
        :param mock_get_statuses: mock_get_statuses
        :param mock_start_manager: mock_start_manager
        :return: None
        """
        mock_get_statuses.return_value.snort_ids_running = False
        constants.CSLE.CSLE_LEVEL_SUBNETMASK_SUFFIX = ".0/24"
        constants.NETWORKING.ETH2 = "eth2"
        constants.NETWORKING.ETH0 = "eth0"
        constants.GRPC_SERVERS.GRPC_OPTIONS = []
        mock_channel = MagicMock()
        mock_insecure_channel.return_value = mock_channel
        mock_stub = MagicMock()
        mock_channel.__enter__.return_value = mock_stub
        SnortIDSController.start_snort_ids(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)
        mock_start_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)
        self.logger.info.assert_any_call(
            f"Snort IDS is not running on "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}, starting it.")
        mock_insecure_channel.assert_called()
        mock_start_snort_ids.assert_called()

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_manager")
    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController."
           "get_snort_idses_monitor_threads_statuses_by_ip_and_port")
    @patch("csle_collector.snort_ids_manager.query_snort_ids_manager.stop_snort_ids")
    @patch("grpc.insecure_channel")
    def test_stop_snort_ids(self, mock_insecure_channel, mock_stop_snort_ids, mock_get_statuses,
                            mock_start_manager) -> None:
        """
        Test utility function for stopping the Snort IDS on a specific IP


        :param mock_insecure_channel: mock_insecure_channel
        :param mock_stop_snort_ids: mock_stop_snort_ids
        :param mock_get_statuses: mock_get_statuse
        :param mock_start_manager: mock_start_manager
        :return: None
        """
        mock_get_statuses.return_value.snort_ids_running = True
        mock_channel = MagicMock()
        mock_insecure_channel.return_value = mock_channel
        mock_stub = MagicMock()
        mock_channel.__enter__.return_value = mock_stub
        SnortIDSController.stop_snort_ids(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_start_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        self.logger.info.assert_any_call(
            f"Snort IDS is running on {self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}, "
            f"stopping it.")
        mock_insecure_channel.assert_called()
        mock_stop_snort_ids.assert_called()

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_manager")
    @patch("time.sleep", return_value=None)
    def test_start_snort_managers(self, mock_sleep, mock_start_manager) -> None:
        """
        Test utility function for starting snort IDS managers

        :param mock_start_manager: mock_start_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        SnortIDSController.start_snort_managers(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_start_manager.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_snort_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test utility function for starting the snort IDS manager on a specific IP

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param mock_sleep: mock_sleep
        :return: None
        """
        emulation_env_config = self.emulation_env_config
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_port = 50051
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_log_dir = "/var/log/snort"
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_log_file = "snort.log"
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_max_workers = 4

        mock_connection = MagicMock()
        emulation_env_config.get_connection.return_value = mock_connection
        mock_execute_ssh_cmd.side_effect = [
            (b"", b"", 0),  # Output for checking if ids_manager is running (not running)
            (b"", b"", 0),  # Output for stopping old background job
            (b"", b"", 0),  # Output for starting the ids_manager
        ]
        SnortIDSController.start_snort_manager(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_connect_admin.assert_called_once_with(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()
        self.logger.info.assert_any_call(
            f"Starting Snort IDS manager on node "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}")

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.stop_snort_manager")
    @patch("time.sleep", return_value=None)
    def test_stop_snort_managers(self, mock_sleep, mock_stop_manager) -> None:
        """
        Test Utility function for stopping snort IDS managers

        :param mock_stop_manager: mock_stop_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        SnortIDSController.stop_snort_managers(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_stop_manager.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_stop_snort_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test utility function for stopping the snort IDS manager on a specific IP

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param mock_sleep: mock_sleep
        :return: None
        """
        emulation_env_config = self.emulation_env_config
        mock_connection = MagicMock()
        emulation_env_config.get_connection.return_value = mock_connection
        mock_execute_ssh_cmd.return_value = (b"", b"", 0)
        SnortIDSController.stop_snort_manager(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_connect_admin.assert_called_once_with(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()
        self.logger.info.assert_any_call(
            f"Stopping Snort IDS manager on node "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}")

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_idses_monitor_thread")
    def test_start_snort_idses_monitor_threads(self, mock_start_monitor_thread) -> None:
        """
        Test a method that sends a request to the SnortIDSManager on every container that runs
        an IDS to start the IDS manager and the monitor thread

        :param mock_start_monitor_thread: mock_start_monitor_thread
        :return: None
        """
        SnortIDSController.start_snort_idses_monitor_threads(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_start_monitor_thread.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        self.logger.info.assert_called_once_with(
            f"Starting Snort IDS monitor thread on IP: "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip} "
            f"({self.emulation_env_config.containers_config.containers[0].get_ips()[0]}, "
            f"{self.emulation_env_config.containers_config.containers[0].get_full_name()})")

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_manager")
    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController."
           "get_snort_idses_monitor_threads_statuses_by_ip_and_port")
    @patch("csle_collector.snort_ids_manager.query_snort_ids_manager.start_snort_ids_monitor")
    @patch("grpc.insecure_channel")
    def test_start_snort_idses_monitor_thread(self, mock_insecure_channel, mock_start_monitor, mock_get_statuses,
                                              mock_start_manager) -> None:
        """
        Test a method that sends a request to the SnortIDSManager on a specific container that runs
        an IDS to start the IDS manager and the monitor thread

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_monitor: mock_start_monitor
        :param mock_get_statuses: mock_get_statuses
        :param mock_start_manager: mock_start_manager
        :return: None
        """
        emulation_env_config = self.emulation_env_config
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_port = 50051
        emulation_env_config.kafka_config.container.get_ips.return_value = ["127.0.0.1"]
        emulation_env_config.kafka_config.kafka_port = 9092
        emulation_env_config.snort_ids_manager_config.time_step_len_seconds = 60
        mock_get_statuses.return_value.monitor_running = False
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        SnortIDSController.start_snort_idses_monitor_thread(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_start_manager.assert_called_once_with(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_get_statuses.assert_called_once_with(
            port=50051, ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)
        self.logger.info.assert_called_once_with(
            f"Snort IDS monitor thread is not running on "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}, starting it.")
        mock_insecure_channel.assert_called_once_with(
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}:50051",
            options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_start_monitor.assert_called()

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.stop_snort_idses_monitor_thread")
    def test_stop_snort_idses_monitor_threadas(self, mock_stop_monitor_thread) -> None:
        """
        Test a method that sends a request to the SnortIDSManager on every container that runs
        an IDS to stop the monitor threads

        :param mock_stop_monitor_thread: mock_stop_monitor_thread
        :return: None
        """
        SnortIDSController.stop_snort_idses_monitor_threads(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_stop_monitor_thread.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_manager")
    @patch("csle_collector.snort_ids_manager.query_snort_ids_manager.stop_snort_ids_monitor")
    @patch("grpc.insecure_channel")
    def test_stop_snort_idses_monitor_thread(self, mock_insecure_channel, mock_stop_monitor, mock_start_manager):
        """
        Test a method that sends a request to the SnortIDSManager on a specific container that runs
        an IDS to stop the monitor threads

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_stop_monitor: mock_stop_monitor
        :param mock_start_manager: mock_start_manager
        :return: None
        """
        emulation_env_config = self.emulation_env_config
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_port = 50051
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        SnortIDSController.stop_snort_idses_monitor_thread(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_start_manager.assert_called_once_with(
            emulation_env_config=emulation_env_config,
            ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        self.logger.info.assert_called_once_with(
            "Stopping the Snort IDS monitor thread on "
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}.")
        mock_insecure_channel.assert_called_once_with(
            f"{self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}:50051",
            options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_stop_monitor.assert_called()

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.start_snort_managers")
    @patch(
        "csle_common.controllers.snort_ids_controller.SnortIDSController."
        "get_snort_idses_monitor_threads_statuses_by_ip_and_port"
    )
    def test_get_snort_idses_monitor_threads_statuses(self, mock_get_statuses, mock_start_managers) -> None:
        """
        Test a method that sends a request to the SnortIDSManager on every container to get the status of the
        IDS monitor thread

        :param mock_get_statuses: mock_get_statuses
        :param mock_start_managers: mock_start_managers
        :return: None
        """
        emulation_env_config = self.emulation_env_config
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_port = 50051
        mock_status = MagicMock()
        mock_get_statuses.return_value = mock_status
        statuses = SnortIDSController.get_snort_idses_monitor_threads_statuses(
            emulation_env_config=emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger, start_if_stopped=True)
        mock_start_managers.assert_called_once_with(
            emulation_env_config=emulation_env_config,
            physical_server_ip=self.emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_get_statuses.assert_called_once_with(
            port=50051, ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)
        assert statuses == [mock_status]

    @patch("grpc.insecure_channel")
    @patch("csle_collector.snort_ids_manager.query_snort_ids_manager.get_snort_ids_monitor_status")
    def test_get_snort_idses_monitor_threads_statuses_by_ip_and_port(self, mock_get_status, mock_insecure_channel) \
            -> None:
        """
        Test a method that sends a request to the SnortIDSManager with a specific port and ip
        to get the status of the IDS monitor thread

        :param mock_get_status: mock_get_status
        :param mock_insecure_channel: mock_insecure_channel
        :return: None
        """
        mock_status = MagicMock()
        mock_get_status.return_value = mock_status
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        port = 50051
        ip = self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip
        status = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(port=port, ip=ip)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        assert status == mock_status

    def test_get_snort_ids_managers_ips(self) -> None:
        """
        Test a method that extracts the IPs of the snort IDS managers in a given emulation

        :return: None
        """
        ips = SnortIDSController.get_snort_ids_managers_ips(emulation_env_config=self.emulation_env_config)
        expected_ips = [self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
                        self.emulation_env_config.containers_config.containers[1].docker_gw_bridge_ip]
        assert ips == expected_ips

    def test_get_snort_idses_managers_ports(self) -> None:
        """
        Test a method that extracts the ports of the snort IDS managers in a given emulation

        :return: None
        """
        ports = SnortIDSController.get_snort_idses_managers_ports(emulation_env_config=self.emulation_env_config)
        expected_ports = [50051, 50051]
        assert ports == expected_ports

    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.get_snort_ids_managers_ips")
    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController.get_snort_idses_managers_ports")
    @patch("csle_common.controllers.snort_ids_controller.SnortIDSController."
           "get_snort_idses_monitor_threads_statuses_by_ip_and_port")
    @patch("csle_common.util.emulation_util.EmulationUtil.physical_ip_match")
    @patch("csle_collector.snort_ids_manager.snort_ids_manager_util.SnortIdsManagerUtil.snort_ids_monitor_dto_empty")
    def test_get_snort_managers_info(
            self, mock_snort_ids_monitor_dto_empty, mock_physical_ip_match, mock_get_statuses, mock_get_ports,
            mock_get_ips) -> None:
        """
        Test the method that extracts the information of the Snort managers for a given emulation

        :param mock_snort_ids_monitor_dto_empty: mock_snort_ids_monitor_dto_empty
        :param mock_physical_ip_match: mock_physical_ip_match
        :param mock_get_statuses: mock_get_statuses
        :param mock_get_ports:mock_get_ports
        :param mock_get_ips: mock_get_ips
        :return: None
        """
        mock_get_ips.return_value = [self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
                                     self.emulation_env_config.containers_config.containers[1].docker_gw_bridge_ip]
        mock_get_ports.return_value = [50051, 50052]
        mock_status = MagicMock()
        mock_get_statuses.side_effect = [mock_status, Exception("Test exception")]
        mock_physical_ip_match.side_effect = [True, False]
        mock_empty_dto = MagicMock()
        mock_snort_ids_monitor_dto_empty.return_value = mock_empty_dto
        emulation_env_config = self.emulation_env_config
        emulation_env_config.snort_ids_manager_config.snort_ids_manager_port = 50051
        emulation_env_config.execution_id = "test_execution_id"
        emulation_env_config.name = "test_emulation_name"
        active_ips = [self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip]
        physical_server_ip = self.emulation_env_config.containers_config.containers[0].physical_server_ip
        SnortIDSController.get_snort_managers_info(emulation_env_config=emulation_env_config, active_ips=active_ips,
                                                   logger=self.logger, physical_server_ip=physical_server_ip)
        mock_get_ips.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_ports.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_statuses.assert_any_call(
            port=50051, ip=self.emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)



from typing import Union, Dict, Any
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class DefaultNetworkFirewallConfig(JSONSerializable):
    """
    DTO representing a default firewall configuration
    """

    def __init__(self, ip: Union[str, None], default_gw: Union[str, None], default_input: str, default_output: str,
                 default_forward: str, network: ContainerNetwork):
        """
        Initializes the DTO

        :param ip: the ip associated to the network
        :param default_gw: the default gateway for the network
        :param default_input: the default input policy for the network
        :param default_output: the default output policy for the network
        :param default_forward: the default forward policy for the network
        :param network: the network configuraiton
        """
        self.ip = ip
        self.default_gw = default_gw
        self.default_input = default_input
        self.default_output = default_output
        self.default_forward = default_forward
        self.network = network

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DefaultNetworkFirewallConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = DefaultNetworkFirewallConfig(
            ip=d["ip"], default_gw=d["default_gw"], default_input=d["default_input"],
            default_output=d["default_output"], default_forward=d["default_forward"],
            network=ContainerNetwork.from_dict(d["network"]))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["default_gw"] = self.default_gw
        d["default_input"] = self.default_input
        d["default_output"] = self.default_output
        d["default_forward"] = self.default_forward
        d["network"] = self.network.to_dict()
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, default_gw:{self.default_gw}, default_input:{self.default_input}, " \
               f"default_output:{self.default_output}, default_forward:{self.default_forward}, network:{self.network}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "DefaultNetworkFirewallConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DefaultNetworkFirewallConfig.from_dict(json.loads(json_str))

    def copy(self) -> "DefaultNetworkFirewallConfig":
        """
        :return: a copy of the DTO
        """
        return DefaultNetworkFirewallConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "DefaultNetworkFirewallConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        if config.ip is not None:
            config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        if config.default_gw is not None:
            config.default_gw = GeneralUtil.replace_first_octet_of_ip(ip=config.default_gw,
                                                                      ip_first_octet=ip_first_octet)
        config.network = config.network.create_execution_config(ip_first_octet=ip_first_octet)
        return config

    @staticmethod
    def schema() -> "DefaultNetworkFirewallConfig":
        """
        :return: get the schema of the DTO
        """
        return DefaultNetworkFirewallConfig(ip="", default_gw="", default_input="", default_output="",
                                            default_forward="", network=ContainerNetwork.schema())



"""
Type of vulnerabilities in an emulation
"""

from enum import IntEnum


class VulnType(IntEnum):
    """
    Enum representing the different vulnerability types for generated containers
    """
    WEAK_PW = 0
    RCE = 1  # remote code execution
    SQL_INJECTION = 2
    PRIVILEGE_ESCALATION = 3



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class User(JSONSerializable):
    """
    DTO class representing a user in an emulation
    """

    def __init__(self, username: str, pw: str, root: bool):
        """
        Initializes the DTO

        :param username: the username of the user
        :param pw: the password of the user
        :param root: whether the user has root access or not
        """
        self.username = username
        self.pw = pw
        self.root = root

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "User":
        """
        Converts a dict representation of the object into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = User(
            username=d["username"],
            pw=d["pw"],
            root=d["root"]
        )
        return obj

    def to_dict(self) -> Dict[str, Union[str, bool]]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, bool]] = {}
        d["username"] = self.username
        d["pw"] = self.pw
        d["root"] = self.root
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"username:{self.username}, pw:{self.pw}, root:{self.root}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "User":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return User.from_dict(json.loads(json_str))

    def copy(self) -> "User":
        """
        :return: a copy of the DTO
        """
        return User.from_dict(self.to_dict())



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class NmapOs(JSONSerializable):
    """
    DTO representing the operating system found with an NMAP scan
    """

    def __init__(self, name: str, vendor: str, osfamily: str, accuracy: int):
        """
        Initializes the DTO object

        :param name: the name of the operating system
        :param vendor: the vendor of the operating system
        :param osfamily: the family of the operating system
        :param accuracy: the accuracy of the OS guess
        """
        self.name = name
        self.vendor = vendor
        self.osfamily = osfamily
        self.accuracy = accuracy

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name:{self.name}, vendor:{self.vendor}, os_family:{self.osfamily}, accuracy:{self.accuracy}"

    @staticmethod
    def get_best_match(os_matches):
        """
        Returns the best matching os

        :param os_matches: list of os matches
        :return: best matching os
        """
        best_accuracy = 0
        best_os = None
        for os in os_matches:
            if os.accuracy > best_accuracy:
                best_os = os
        return best_os

    def to_dict(self) -> Dict[str, Union[str, int]]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int]] = {}
        d["name"] = self.name
        d["vendor"] = self.vendor
        d["osfamily"] = self.osfamily
        d["accuracy"] = self.accuracy
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapOs":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapOs(name=d["name"], vendor=d["vendor"], osfamily=d["osfamily"], accuracy=d["accuracy"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapOs":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapOs.from_dict(json.loads(json_str))



from csle_common.util.read_emulation_statistics_util import ReadEmulationStatisticsUtil
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_alert_counters import SnortIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
import csle_collector.constants.constants as constants
import numpy as np


class TestReadEmulationUtilSuite:
    """
    Test suite for read_emulation_util
    """

    def test_average_host_metrics(self) -> None:
        """
        Test the method that computes the average metrics from a list of host metrics

        :return: None
        """
        metrics1 = HostMetrics(
            num_logged_in_users=5, num_failed_login_attempts=2, num_open_connections=10, num_login_events=3,
            num_processes=50, num_users=5)
        metrics2 = HostMetrics(
            num_logged_in_users=7, num_failed_login_attempts=4, num_open_connections=14, num_login_events=5,
            num_processes=60, num_users=6)
        metrics3 = HostMetrics(
            num_logged_in_users=6, num_failed_login_attempts=3, num_open_connections=12, num_login_events=4,
            num_processes=55, num_users=7)
        average_metrics = ReadEmulationStatisticsUtil.average_host_metrics([metrics1, metrics2, metrics3])
        assert average_metrics.num_logged_in_users == 18
        assert average_metrics.num_failed_login_attempts == 9
        assert average_metrics.num_open_connections == 36
        assert average_metrics.num_login_events == 12
        assert average_metrics.num_processes == 165
        assert average_metrics.num_users == 18

    def test_average_ossec_metrics(self) -> None:
        """
        Test the method that computes the average metrics from a list of OSSEC metrics

        :return: None
        """
        metrics1 = OSSECIdsAlertCounters()
        metrics1.level_alerts = list(np.zeros(16))
        metrics1.group_alerts = list(np.zeros(len(set(constants.OSSEC.OSSEC_IDS_ALERT_GROUP_ID.values()))))
        metrics1.severe_alerts = 5
        metrics1.warning_alerts = 3
        metrics1.total_alerts = 10
        metrics1.alerts_weighted_by_level = 20

        metrics2 = OSSECIdsAlertCounters()
        metrics2.level_alerts = list(np.zeros(16))
        metrics2.group_alerts = list(np.zeros(len(set(constants.OSSEC.OSSEC_IDS_ALERT_GROUP_ID.values()))))
        metrics2.severe_alerts = 10
        metrics2.warning_alerts = 6
        metrics2.total_alerts = 20
        metrics2.alerts_weighted_by_level = 40
        average_metrics = ReadEmulationStatisticsUtil.average_ossec_metrics([metrics1, metrics2])
        assert average_metrics.level_alerts == [0] * 16
        assert average_metrics.group_alerts == [0] * len(set(constants.OSSEC.OSSEC_IDS_ALERT_GROUP_ID.values()))
        assert average_metrics.severe_alerts == 15
        assert average_metrics.warning_alerts == 9
        assert average_metrics.total_alerts == 30
        assert average_metrics.alerts_weighted_by_level == 60

    def test_average_snort_metrics(self) -> None:
        """
        Test the method that computes the average metrics from a list of Snort metrics

        :return: None
        """
        metrics1 = SnortIdsAlertCounters()
        metrics1.priority_alerts = [0] * 4
        metrics1.class_alerts = []
        for i in range(len(set(constants.SNORT_IDS_ROUTER.SNORT_ALERT_IDS_ID.values()))):
            metrics1.class_alerts.append(0)
        metrics1.severe_alerts = 5
        metrics1.warning_alerts = 3
        metrics1.total_alerts = 10
        metrics1.alerts_weighted_by_priority = 20

        metrics2 = SnortIdsAlertCounters()
        metrics2.priority_alerts = [0] * 4
        metrics2.class_alerts = []
        for i in range(len(set(constants.SNORT_IDS_ROUTER.SNORT_ALERT_IDS_ID.values()))):
            metrics2.class_alerts.append(0)
        metrics2.severe_alerts = 10
        metrics2.warning_alerts = 6
        metrics2.total_alerts = 20
        metrics2.alerts_weighted_by_priority = 40
        average_metrics = ReadEmulationStatisticsUtil.average_snort_metrics([metrics1, metrics2])
        assert average_metrics.priority_alerts == [0] * 4
        assert average_metrics.class_alerts == [0] * len(set(constants.SNORT_IDS_ROUTER.SNORT_ALERT_IDS_ID.values()))
        assert average_metrics.severe_alerts == 15
        assert average_metrics.warning_alerts == 9
        assert average_metrics.total_alerts == 30
        assert average_metrics.alerts_weighted_by_priority == 60

    def test_average_snort_rule_metrics(self) -> None:
        """
        Test the method that computes the average metrics from a list of Snort rule metrics

        :return: None
        """
        metrics1 = SnortIdsRuleCounters()
        metrics1.rule_alerts = {}
        metrics2 = SnortIdsAlertCounters()
        metrics2.rule_alerts = {}
        average_metrics = ReadEmulationStatisticsUtil.average_snort_rule_metrics([metrics1, metrics2])
        assert average_metrics.rule_alerts == {}



from typing import List
import threading
import time
from confluent_kafka import Consumer, KafkaError, KafkaException
import csle_collector.constants.constants as collector_constants
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
from csle_common.logging.log import Logger


class AggregatedSnortIdsRuleLogConsumerThread(threading.Thread):
    """
    Thread that polls the Snort IDS log to get the latest rule-based metrics
    """

    def __init__(self, kafka_server_ip: str, kafka_port: int, snort_ids_rule_counters: SnortIdsRuleCounters,
                 auto_offset_reset: str = "latest") -> None:
        """
        Initializes the thread

        :param kafka_server_ip: the ip of the kafka server
        :param kafka_port: the port of the kafka server
        :param snort_ids_rule_counters: the rule counters to update
        :param auto_offset_reset: the offset for kafka to start reading from
        """
        threading.Thread.__init__(self)
        self.running = True
        self.kafka_server_ip = kafka_server_ip
        self.kafka_port = kafka_port
        self.snort_ids_rule_counters = snort_ids_rule_counters
        self.snort_ids_rule_counters_list: List[SnortIdsRuleCounters] = []
        self.ts = time.time()
        self.kafka_conf = {
            collector_constants.KAFKA.BOOTSTRAP_SERVERS_PROPERTY: f"{self.kafka_server_ip}:{self.kafka_port}",
            collector_constants.KAFKA.GROUP_ID_PROPERTY: f"ids_log_consumer_thread_{self.ts}",
            collector_constants.KAFKA.AUTO_OFFSET_RESET_PROPERTY: auto_offset_reset}
        self.consumer = Consumer(**self.kafka_conf)
        self.consumer.subscribe([collector_constants.KAFKA_CONFIG.SNORT_IDS_RULE_LOG_TOPIC_NAME])

    def run(self) -> None:
        """
        Runs the thread

        :return: None
        """
        while self.running:
            msg = self.consumer.poll(timeout=5.0)
            if msg is not None:
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        Logger.__call__().get_logger.warning(
                            f"reached end of partition: {msg.topic(), msg.partition(), msg.offset()}")
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    self.snort_ids_rule_counters.update_with_kafka_record(record=msg.value().decode())
                    self.snort_ids_rule_counters_list.append(self.snort_ids_rule_counters.copy())
        self.consumer.close()

    def get_aggregated_ids_rule_counters(self) -> SnortIdsRuleCounters:
        """
        :return: aggregated alert counters from the list
        """
        import logging
        logging.info(len(self.snort_ids_rule_counters_list))
        if len(self.snort_ids_rule_counters_list) == 0:
            return self.snort_ids_rule_counters.copy()
        if len(self.snort_ids_rule_counters_list) == 1:
            return self.snort_ids_rule_counters_list[0].copy()
        rule_counters = self.snort_ids_rule_counters_list[0].copy()
        for j in range(1, len(self.snort_ids_rule_counters_list)):
            rule_counters.add(self.snort_ids_rule_counters_list[j].copy())
        return rule_counters



import threading
import time
from csle_common.tunneling.forward_ssh_server import ForwardSSHServer
from csle_common.tunneling.forward_ssh_controller import ForwardSSHHandler
from typing import Dict, Any


class ForwardTunnelThread(threading.Thread):
    """
    Thread that starts up a SSH tunnel that forwards a local port to a remote machine
    """

    def __init__(self, local_port: int, remote_host: str, remote_port: int,
                 transport, tunnels_dict: Dict[str, Any] = {}) -> None:
        """
        Initializes the thread

        :param local_port: the local port for port-forwarding
        :param remote_host: the remote host
        :param remote_port: the remote port
        :param transport: the transport protocol
        :param tunnels_dict: the tunnels dict for garbage collection
        """
        super().__init__()
        self.local_port = local_port
        self.remote_host = remote_host
        self.transport = transport
        self.remote_port = remote_port
        self.forward_server = ForwardSSHServer(("", local_port), ForwardSSHHandler)
        self.forward_server.ssh_transport = self.transport
        self.forward_server.chain_host = self.remote_host
        self.forward_server.chain_port = self.remote_port
        self.forward_server.tunnels_dict = tunnels_dict
        self.daemon = True

    def run(self) -> None:
        """
        Starts the server

        :return:
        """
        self.forward_server.serve_forever()

    def shutdown(self) -> None:
        """
        Shutsdown the server

        :return: None
        """
        self.forward_server.shutdown()
        time.sleep(0.5)  # wait for server to shutdown



import logging
from typing import List
import grpc
import time
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.elk_managers_info import ELKManagersInfo
import csle_common.constants.constants as constants
import csle_collector.elk_manager.elk_manager_pb2_grpc
import csle_collector.elk_manager.elk_manager_pb2
import csle_collector.elk_manager.query_elk_manager
import csle_collector.elk_manager.elk_manager_util
import csle_collector.constants.constants as collector_constants
from csle_common.util.emulation_util import EmulationUtil


class ELKController:
    """
    Controller managing API calls to ELK-stacks running inside emulations
    """

    @staticmethod
    def start_elk_manager(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Utility method for starting the ELK manager

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if elk_manager is already running
        status = None
        not_running = False
        try:
            status = ELKController.get_elk_status_by_port_and_ip(
                ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                port=emulation_env_config.elk_config.elk_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = (f"Elasticrunning: {status.elasticRunning}, kibanaRunning: {status.kibanaRunning}, "
                          f"logstashRunning: {status.logstashRunning}")
        if not_running:
            logger.info(f"Starting elk manager on node: "
                        f"{emulation_env_config.elk_config.container.docker_gw_bridge_ip}")

            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                        ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                        create_producer=False)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.ELK_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd,
                conn=emulation_env_config.get_connection(
                    ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip))

            # Start the elk_manager
            cmd = constants.COMMANDS.START_ELK_MANAGER.format(
                emulation_env_config.elk_config.elk_manager_port, emulation_env_config.elk_config.elk_manager_log_dir,
                emulation_env_config.elk_config.elk_manager_log_file,
                emulation_env_config.elk_config.elk_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd,
                conn=emulation_env_config.get_connection(
                    ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip))
            time.sleep(2)
        else:
            logger.info(f"ELK manager already running on node: "
                        f"{emulation_env_config.elk_config.container.docker_gw_bridge_ip}. Status: {status_str}")

    @staticmethod
    def stop_elk_manager(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Utility method for stopping the ELK manager

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                    ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                    create_producer=False)

        logger.info(f"Stopping elk manager on node: "
                    f"{emulation_env_config.elk_config.container.docker_gw_bridge_ip}")

        # Stop background job
        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.ELK_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(
            cmd=cmd,
            conn=emulation_env_config.get_connection(
                ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip))

        time.sleep(2)

    @staticmethod
    def get_elk_status(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for querying the ELKManager about the status of the ELK stack

        :param emulation_env_config: the emulation config
        :return: an ELKDTO with the status of the server
        """
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)
        elk_dto = ELKController.get_elk_status_by_port_and_ip(
            ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
            port=emulation_env_config.elk_config.elk_manager_port)
        return elk_dto

    @staticmethod
    def get_elk_status_by_port_and_ip(ip: str, port: int, timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for querying the ElkManager about the status of the ELK stack

        :param ip: the ip where the ElkManager is running
        :param port: the port the ELkManager is listening to
        :param timeout: the timeout of the GRPC query
        :return: an ELKDTO with the status of the server
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.get_elk_status(stub, timeout=timeout)
            return elk_dto

    @staticmethod
    def stop_elk_stack(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to stop the ELK server

        :param emulation_env_config: the emulation env config
        :return: a ELKDTO with the status of the server
        """
        logger.info(
            f"Stopping ELK stack on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.stop_elk(stub)
            return elk_dto

    @staticmethod
    def start_elk_stack(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start the ELK server

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        if emulation_env_config.elk_config.container.physical_host_ip != physical_server_ip:
            return
        logger.info(
            f"Starting ELK stack on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.start_elk(stub)
            return elk_dto

    @staticmethod
    def start_elastic(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start elasticsearch

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        logger.info(
            f"Starting elasticsearch on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.start_elastic(stub)
            return elk_dto

    @staticmethod
    def start_kibana(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start kibana

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        logger.info(
            f"Starting kibana on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.start_kibana(stub)
            return elk_dto

    @staticmethod
    def start_logstash(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start Logstash

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        logger.info(
            f"Starting logstash on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.start_logstash(stub)
            return elk_dto

    @staticmethod
    def stop_elastic(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start elasticsearch

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        logger.info(
            f"Stopping elasticsearch on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.stop_elastic(stub)
            return elk_dto

    @staticmethod
    def stop_kibana(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start kibana

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        logger.info(
            f"Stopping kibana on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.stop_kibana(stub)
            return elk_dto

    @staticmethod
    def stop_logstash(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.elk_manager.elk_manager_pb2.ElkDTO:
        """
        Method for requesting the ELKManager to start Logstash

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an ELKDTO with the status of the server
        """
        logger.info(
            f"Stopping logstash on container: {emulation_env_config.elk_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.elk_config.container.get_ips()[0]}, "
            f"{emulation_env_config.elk_config.container.get_full_name()})")
        ELKController.start_elk_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.elk_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.elk_config.elk_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub(channel)
            elk_dto = csle_collector.elk_manager.query_elk_manager.stop_logstash(stub)
            return elk_dto

    @staticmethod
    def get_elk_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPS of the ELK managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        return [emulation_env_config.elk_config.container.docker_gw_bridge_ip]

    @staticmethod
    def get_elk_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the Elk managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        return [emulation_env_config.elk_config.elk_manager_port]

    @staticmethod
    def get_elk_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                              logger: logging.Logger, physical_host_ip: str) -> ELKManagersInfo:
        """
        Extracts the information of the ELK managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_host_ip: the physical host ip
        :param logger: the logger to use for logging
        :return: a DTO with the status of the ELK managers
        """
        elk_managers_ips = ELKController.get_elk_managers_ips(emulation_env_config=emulation_env_config)
        elk_managers_ports = ELKController.get_elk_managers_ports(emulation_env_config=emulation_env_config)
        elk_managers_statuses = []
        elk_managers_running = []
        if emulation_env_config.elk_config.container.physical_host_ip == physical_host_ip:
            for ip in elk_managers_ips:
                if ip not in active_ips:
                    continue
                status = None
                try:
                    status = ELKController.get_elk_status_by_port_and_ip(
                        port=emulation_env_config.elk_config.elk_manager_port, ip=ip)
                    running = True
                except Exception as e:
                    running = False
                    logger.debug(
                        f"Could not fetch Elk manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
                if status is not None:
                    elk_managers_statuses.append(status)
                else:
                    elk_managers_statuses.append(
                        csle_collector.elk_manager.elk_manager_util.ElkManagerUtil.elk_dto_empty())
                elk_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        elk_manager_info_dto = ELKManagersInfo(
            elk_managers_running=elk_managers_running, ips=elk_managers_ips, execution_id=execution_id,
            emulation_name=emulation_name, elk_managers_statuses=elk_managers_statuses, ports=elk_managers_ports,
            physical_server_ip=emulation_env_config.elk_config.container.physical_host_ip)
        return elk_manager_info_dto



from typing import List, Tuple, Union, Set, Any
import csle_common.constants.constants as constants
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_common.dao.emulation_observation.common.emulation_port_observation_state \
    import EmulationPortObservationState
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state \
    import EmulationVulnerabilityObservationState
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_observation.attacker.emulation_attacker_observation_state \
    import EmulationAttackerObservationState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_config import EmulationAttackerActionConfig


class EnvDynamicsUtil:
    """
    Class containing common utilities that are used both in simulation-mode and in emulation-mode.
    """

    @staticmethod
    def merge_complete_obs_state(old_obs_state: EmulationAttackerObservationState,
                                 new_obs_state: EmulationAttackerObservationState,
                                 emulation_env_config: EmulationEnvConfig) -> EmulationAttackerObservationState:
        """
        Merges an old observation state with a new one

        :param old_obs_state: the old observation state
        :param new_obs_state: the new observation state
        :param emulation_env_config: the emulation environment configuration
        :return: the merged observation state
        """
        merged_obs_state = old_obs_state.copy()
        merged_obs_state.catched_flags = max(old_obs_state.catched_flags, new_obs_state.catched_flags)
        merged_obs_state.machines = EnvDynamicsUtil.merge_new_obs_with_old(
            old_obs_state.machines, new_obs_state.machines, emulation_env_config=emulation_env_config, action=None)
        merged_obs_state.agent_reachable = old_obs_state.agent_reachable.union(new_obs_state.agent_reachable)
        return merged_obs_state

    @staticmethod
    def merge_new_obs_with_old(old_machines_obs: List[EmulationAttackerMachineObservationState],
                               new_machines_obs: List[EmulationAttackerMachineObservationState],
                               emulation_env_config: EmulationEnvConfig,
                               action: Union[EmulationAttackerAction, None]) \
            -> List[EmulationAttackerMachineObservationState]:
        """
        Helper function for merging an old network observation with new information collected

        :param old_machines_obs: the list of old machine observations
        :param new_machines_obs: the list of newly collected information
        :param emulation_env_config: environment config
        :param action: the action
        :return: the merged machine observations
        """
        if action is None:
            raise ValueError("action is None")
        new_machines_obs = EnvDynamicsUtil.merge_duplicate_machines(machines=new_machines_obs, action=action)
        attacker_machine_observations = []
        # Add updated machines to merged state
        for n_m in new_machines_obs:
            if emulation_env_config.containers_config.agent_ip in n_m.ips:
                continue
            merged_m = n_m
            for i, o_m in enumerate(old_machines_obs):
                if n_m.ips_match(o_m.ips):
                    merged_m = EnvDynamicsUtil.merge_new_machine_obs_with_old_machine_obs(o_m, n_m, action)
            attacker_machine_observations.append(merged_m)

        # Add old machines to merged state
        for o_m in old_machines_obs:
            exists = False
            for m_m in attacker_machine_observations:
                if o_m.ips_match(m_m.ips):
                    exists = True
            if not exists:
                attacker_machine_observations.append(o_m)

        return attacker_machine_observations

    @staticmethod
    def merge_new_machine_obs_with_old_machine_obs(o_m: EmulationAttackerMachineObservationState,
                                                   n_m: EmulationAttackerMachineObservationState,
                                                   action: EmulationAttackerAction) \
            -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation with new information collected

        :param o_m: old machine observation
        :param n_m: newly collected machine information
        :return: the merged attacker machine observation state
        """
        
        merged_ports, num_new_ports_found = EnvDynamicsUtil.merge_ports(o_m.ports, n_m.ports, acc=True)
        n_m.ports = merged_ports
        merged_os, num_new_os_found = EnvDynamicsUtil.merge_os(o_m.os, n_m.os)
        n_m.os = merged_os
        merged_cve_vulnerabilities, num_new_cve_vuln_found = EnvDynamicsUtil.merge_vulnerabilities(o_m.cve_vulns,
                                                                                                   n_m.cve_vulns)
        n_m.cve_vulns = merged_cve_vulnerabilities
        merged_osvdb_vulnerabilities, num_new_osvdb_vuln_found = EnvDynamicsUtil.merge_vulnerabilities(
            o_m.osvdb_vulns, n_m.osvdb_vulns)
        n_m.osvdb_vulns = merged_osvdb_vulnerabilities
        n_m, new_shell_access = EnvDynamicsUtil.merge_shell_access(o_m, n_m)
        n_m, num_new_logged_in = EnvDynamicsUtil.merge_logged_in(o_m, n_m)
        n_m, new_root = EnvDynamicsUtil.merge_root(o_m, n_m)
        n_m, new_flag_pts = EnvDynamicsUtil.merge_flags(o_m, n_m)
        n_m = EnvDynamicsUtil.merge_connections(o_m, n_m)
        n_m = EnvDynamicsUtil.merge_filesystem_scanned(o_m, n_m, new_root)
        n_m = EnvDynamicsUtil.merge_untried_credentials(o_m, n_m, action)
        n_m = EnvDynamicsUtil.merge_trace(o_m, n_m)
        n_m = EnvDynamicsUtil.merge_exploit_tried(o_m, n_m)
        n_m = EnvDynamicsUtil.merge_backdoor_tried(o_m, n_m)
        n_m = EnvDynamicsUtil.merge_tools_tried(o_m, n_m)
        n_m, num_new_tools_installed = EnvDynamicsUtil.merge_tools_installed(o_m, n_m)
        n_m, num_new_backdoors_installed = EnvDynamicsUtil.merge_backdoor_installed(o_m, n_m)
        n_m = EnvDynamicsUtil.merge_reachable(o_m=o_m, n_m=n_m)
        n_m = EnvDynamicsUtil.merge_backdoor_credentials(o_m=o_m, n_m=n_m)

        return n_m

    @staticmethod
    def merge_os(o_os: str, n_os: str) -> Tuple[str, int]:
        """
        Helper function for merging an old machine observation OS with new information collected

        :param o_os: the old OS
        :param n_os: the newly observed OS
        :return: the merged os, 1 if it was a newly detected OS, otherwise 0
        """
        merged_os = n_os
        if n_os == "unknown" and o_os != "unknown":
            merged_os = o_os
        new_os = 0 if merged_os == o_os else 1
        return merged_os, new_os

    @staticmethod
    def merge_shell_access(o_m: EmulationAttackerMachineObservationState,
                           n_m: EmulationAttackerMachineObservationState) -> \
            Tuple[EmulationAttackerMachineObservationState, int]:
        """
        Helper function for merging an old machine observation shell access with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with update shell-access parameters, 1 if new access otherwise 0
        """
        new_access = 0
        if not o_m.shell_access and n_m.shell_access:
            new_access = 1
            merged_credentials = o_m.shell_access_credentials
            for cr in n_m.shell_access_credentials:
                duplicate = False
                for cr_2 in o_m.shell_access_credentials:
                    if cr == cr_2:
                        duplicate = True
                if not duplicate:
                    merged_credentials.append(cr)
            n_m.shell_access_credentials = merged_credentials
        if not n_m.shell_access:
            n_m.shell_access = o_m.shell_access
            n_m.shell_access_credentials = o_m.shell_access_credentials
        return n_m, new_access

    @staticmethod
    def merge_reachable(o_m: EmulationAttackerMachineObservationState,
                        n_m: EmulationAttackerMachineObservationState) \
            -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation reachable nodes with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated reachable nodes
        """
        n_m.reachable = n_m.reachable.union(o_m.reachable)
        return n_m

    @staticmethod
    def merge_backdoor_credentials(o_m: EmulationAttackerMachineObservationState,
                                   n_m: EmulationAttackerMachineObservationState) \
            -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation backdoor credentials nodes with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated backdoor credentials nodes
        """
        n_m.backdoor_credentials = list(set(n_m.backdoor_credentials + o_m.backdoor_credentials))
        return n_m

    @staticmethod
    def merge_logged_in(o_m: EmulationAttackerMachineObservationState,
                        n_m: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, int]:
        """
        Helper function for merging an old machine observation logged in with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated logged-in flag, num_new_login
        """
        num_new_logged_in = 0
        if not o_m.logged_in and n_m.logged_in:
            num_new_logged_in = 1
        if not n_m.logged_in:
            n_m.logged_in = o_m.logged_in
        return n_m, num_new_logged_in

    @staticmethod
    def merge_tools_installed(o_m: EmulationAttackerMachineObservationState,
                              n_m: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, int]:
        """
        Helper function for merging an old machine observation "tools installed flag" with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated tools_installed flag, num_new_tools_installed
        """
        num_new_tools_installed = 0
        if not o_m.tools_installed and n_m.tools_installed:
            num_new_tools_installed = 1
        if not n_m.tools_installed:
            n_m.tools_installed = o_m.tools_installed
        return n_m, num_new_tools_installed

    @staticmethod
    def merge_backdoor_installed(o_m: EmulationAttackerMachineObservationState,
                                 n_m: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, int]:
        """
        Helper function for merging an old machine observation "backdoor installed" flag with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated backdoor_installed flag, num_new_backdoor_installed
        """
        num_new_backdoor_installed = 0
        if not o_m.backdoor_installed and n_m.backdoor_installed:
            num_new_backdoor_installed = 1
        if not n_m.backdoor_installed:
            n_m.backdoor_installed = o_m.backdoor_installed
        return n_m, num_new_backdoor_installed

    @staticmethod
    def merge_backdoor_tried(o_m: EmulationAttackerMachineObservationState,
                             n_m: EmulationAttackerMachineObservationState) \
            -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation "backdoor tried" flag with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated backdoor-tried flag
        """
        if not n_m.backdoor_tried:
            n_m.backdoor_tried = o_m.backdoor_tried
        return n_m

    @staticmethod
    def merge_tools_tried(o_m: EmulationAttackerMachineObservationState,
                          n_m: EmulationAttackerMachineObservationState) -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation "tools tried" flag with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated tools-installed-tried flag
        """
        if not n_m.install_tools_tried:
            n_m.install_tools_tried = o_m.install_tools_tried
        return n_m

    @staticmethod
    def merge_root(o_m: EmulationAttackerMachineObservationState, n_m: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, int]:
        """
        Helper function for merging an old machine observation root with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated root flag, 1 if new root otherwise 0
        """
        new_root = 0
        if not o_m.root and n_m.root:
            new_root = 1
        if not n_m.root:
            n_m.root = o_m.root
        return n_m, new_root

    @staticmethod
    def merge_flags(o_m: EmulationAttackerMachineObservationState, n_m: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, int]:
        """
        Helper function for merging an old machine observation flags with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated flags, number of new flag points
        """
        new_flag_points = 0
        for flag in n_m.flags_found:
            if flag not in o_m.flags_found:
                new_flag_points += flag.score
        n_m.flags_found = n_m.flags_found.union(o_m.flags_found)
        return n_m, new_flag_points

    @staticmethod
    def merge_connections(o_m: EmulationAttackerMachineObservationState,
                          n_m: EmulationAttackerMachineObservationState) -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation shell-connections with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated connections
        """
        n_m.ssh_connections = list(set(n_m.ssh_connections + o_m.ssh_connections))
        n_m.telnet_connections = list(set(n_m.telnet_connections + o_m.telnet_connections))
        n_m.ftp_connections = list(set(n_m.ftp_connections + o_m.ftp_connections))
        return n_m

    @staticmethod
    def merge_filesystem_scanned(o_m: EmulationAttackerMachineObservationState,
                                 n_m: EmulationAttackerMachineObservationState,
                                 new_root: int) -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation file-system-scanned-flag with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :param new_root: number of new root connections
        :return: the merged machine observation with updated filesystem-scanned flag
        """
        if not n_m.filesystem_searched:
            n_m.filesystem_searched = o_m.filesystem_searched
        if new_root > 0:
            n_m.filesystem_searched = False
        return n_m

    @staticmethod
    def merge_untried_credentials(o_m: EmulationAttackerMachineObservationState,
                                  n_m: EmulationAttackerMachineObservationState,
                                  action: EmulationAttackerAction) -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation untried-credentials-flag with new information collected

        :param o_m: the old machine observation
        :param n_m: the new machine observation
        :param action: the action that was done to get n_m
        :return: the merged machine observation with updated untried-credentials flag
        """
        if action is not None and action.id == EmulationAttackerActionId.NETWORK_SERVICE_LOGIN and n_m.shell_access:
            return n_m
        else:
            if not n_m.untried_credentials:
                n_m.untried_credentials = o_m.untried_credentials
        return n_m

    @staticmethod
    def merge_exploit_tried(o_m: EmulationAttackerMachineObservationState,
                            n_m: EmulationAttackerMachineObservationState) -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation tried_brute_flags with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated brute-tried flags
        """
        if not n_m.telnet_brute_tried:
            n_m.telnet_brute_tried = o_m.telnet_brute_tried
        if not n_m.ssh_brute_tried:
            n_m.ssh_brute_tried = o_m.ssh_brute_tried
        if not n_m.ftp_brute_tried:
            n_m.ftp_brute_tried = o_m.ftp_brute_tried
        if not n_m.cassandra_brute_tried:
            n_m.cassandra_brute_tried = o_m.cassandra_brute_tried
        if not n_m.irc_brute_tried:
            n_m.irc_brute_tried = o_m.irc_brute_tried
        if not n_m.mongo_brute_tried:
            n_m.mongo_brute_tried = o_m.mongo_brute_tried
        if not n_m.mysql_brute_tried:
            n_m.mysql_brute_tried = o_m.mysql_brute_tried
        if not n_m.smtp_brute_tried:
            n_m.smtp_brute_tried = o_m.smtp_brute_tried
        if not n_m.postgres_brute_tried:
            n_m.postgres_brute_tried = o_m.postgres_brute_tried
        if not n_m.sambacry_tried:
            n_m.sambacry_tried = o_m.sambacry_tried
        if not n_m.shellshock_tried:
            n_m.shellshock_tried = o_m.shellshock_tried
        if not n_m.dvwa_sql_injection_tried:
            n_m.dvwa_sql_injection_tried = o_m.dvwa_sql_injection_tried
        if not n_m.cve_2015_3306_tried:
            n_m.cve_2015_3306_tried = o_m.cve_2015_3306_tried
        if not n_m.cve_2015_1427_tried:
            n_m.cve_2015_1427_tried = o_m.cve_2015_1427_tried
        if not n_m.cve_2016_10033_tried:
            n_m.cve_2016_10033_tried = o_m.cve_2016_10033_tried
        if not n_m.cve_2010_0426_tried:
            n_m.cve_2010_0426_tried = o_m.cve_2010_0426_tried
        if not n_m.cve_2015_5602_tried:
            n_m.cve_2015_5602_tried = o_m.cve_2015_5602_tried
        return n_m

    @staticmethod
    def merge_trace(o_m: EmulationAttackerMachineObservationState, n_m: EmulationAttackerMachineObservationState) \
            -> EmulationAttackerMachineObservationState:
        """
        Helper function for merging an old machine observation trace with new information collected

        :param o_os: the old machine observation
        :param n_os: the new machine observation
        :return: the merged machine observation with updated trace
        """
        if n_m.trace is None:
            n_m.trace = o_m.trace
        return n_m

    @staticmethod
    def merge_ports(o_ports: List[EmulationPortObservationState], n_ports: List[EmulationPortObservationState],
                    acc: bool = True) -> Tuple[List[EmulationPortObservationState], int]:
        """
        Helper function for merging two port lists

        :param o_ports: old list of ports
        :param n_ports: new list of ports
        :param acc: if true, accumulate port observations rather than overwrite
        :return: the merged port list, number of new ports found
        """
        num_new_ports_found = 0
        if n_ports is None or len(n_ports) == 0:
            return o_ports, num_new_ports_found
        merged_ports = n_ports
        for m_p in merged_ports:
            exist = False
            for o_p in o_ports:
                if o_p.port == m_p.port and o_p.protocol == m_p.protocol and o_p.service == m_p.service \
                        and o_p.open == m_p.open:
                    exist = True
            if not exist:
                num_new_ports_found += 1
        if acc:
            for o_p in o_ports:
                exist = False
                for m_p in merged_ports:
                    if o_p.port == m_p.port and o_p.protocol == m_p.protocol:
                        exist = True
                if not exist:
                    merged_ports.append(o_p)
        return merged_ports, num_new_ports_found

    @staticmethod
    def merge_vulnerabilities(o_vuln: List[EmulationVulnerabilityObservationState],
                              n_vuln: List[EmulationVulnerabilityObservationState],
                              acc: bool = True) -> Tuple[List[EmulationVulnerabilityObservationState], int]:
        """
        Helper function for merging two vulnerability lists lists

        :param o_vuln: old list of vulnerabilities
        :param n_vuln: new list of vulnerabilities
        :param acc: if true, accumulate vulnerability observations rather than overwrite
        :return: the merged vulnerability list, number of new vulnerabilities detected
        """
        num_new_vuln_found = 0
        if n_vuln is None or len(n_vuln) == 0:
            return o_vuln, num_new_vuln_found
        merged_vuln = n_vuln
        for m_v in merged_vuln:
            exist = False
            for o_v in o_vuln:
                if o_v.name == m_v.name:
                    exist = True
            if not exist:
                num_new_vuln_found += 1
        if acc:
            for o_v in o_vuln:
                exist = False
                for m_v in merged_vuln:
                    if o_v.name == m_v.name:
                        exist = True
                if not exist:
                    merged_vuln.append(o_v)
        return merged_vuln, num_new_vuln_found

    @staticmethod
    def is_all_flags_collected(s: EmulationEnvState, emulation_env_config: EmulationEnvConfig) -> bool:
        """
        Checks if all flags are collected (then episode is done)

        :param s: current state
        :param emulation_env_config: environment config
        :return: True if all flags are collected otherwise false
        """
        total_flags = set()
        for flag in emulation_env_config.flags_config.node_flag_configs:
            total_flags.add(flag)
        found_flags: Set[Any] = set()
        if s.attacker_obs_state is None:
            raise ValueError("EmlationAttackerObservationState is None")
        for node in s.attacker_obs_state.machines:
            found_flags = found_flags.union(node.flags_found)
        return total_flags == found_flags

    @staticmethod
    def check_if_telnet_connection_is_alive(conn) -> bool:
        """
        Utility function to check whether a Telnet connection is alive or not

        :param conn: the connection to check
        :return: true or false
        """
        return True

    @staticmethod
    def check_if_ftp_connection_is_alive(conn) -> bool:
        """
        Utilitty function to check whether a FTP connection is alive or not

        :param conn: the connection to check
        :return: true or false
        """
        return True

    @staticmethod
    def exploit_tried_flags(a: EmulationAttackerAction, m_obs: EmulationAttackerMachineObservationState):
        """
        Utility function for updating exploit-tried flags

        :param a: the action
        :param m_obs: the observation to update
        :return: the updated observation
        """
        # Subnet scans sometimes fail due to stagnation, https://github.com/nmap/nmap/issues/1321
        if (a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.ftp_brute_tried = True
        elif (a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.ssh_brute_tried = True
        elif (a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.telnet_brute_tried = True
        elif (a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.irc_brute_tried = True
        elif (a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.postgres_brute_tried = True
        elif (a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.smtp_brute_tried = True
        elif (a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.mysql_brute_tried = True
        elif (a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials)\
                or a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.mongo_brute_tried = True
        elif (a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL and m_obs.untried_credentials) \
                or a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST:
            m_obs.cassandra_brute_tried = True
        elif a.id == EmulationAttackerActionId.SAMBACRY_EXPLOIT:
            m_obs.sambacry_tried = True
        elif a.id == EmulationAttackerActionId.SHELLSHOCK_EXPLOIT:
            m_obs.shellshock_tried = True
        elif a.id == EmulationAttackerActionId.DVWA_SQL_INJECTION:
            m_obs.dvwa_sql_injection_tried = True
        elif a.id == EmulationAttackerActionId.CVE_2015_3306_EXPLOIT:
            m_obs.cve_2015_3306_tried = True
        elif a.id == EmulationAttackerActionId.CVE_2015_1427_EXPLOIT:
            m_obs.cve_2015_1427_tried = True
        elif a.id == EmulationAttackerActionId.CVE_2016_10033_EXPLOIT:
            m_obs.cve_2016_10033_tried = True
        elif a.id == EmulationAttackerActionId.CVE_2010_0426_PRIV_ESC:
            m_obs.cve_2010_0426_tried = True
        elif a.id == EmulationAttackerActionId.CVE_2015_5602_PRIV_ESC:
            m_obs.cve_2015_5602_tried = True
        return m_obs

    @staticmethod
    def ssh_backdoor_tried_flags(a: EmulationAttackerAction, m_obs: EmulationAttackerMachineObservationState):
        """
        Utility function for updating install-backdoor flag

        :param a: the action
        :param m_obs: the observation to update
        :return: the updated observation
        """
        if a.id == EmulationAttackerActionId.SSH_BACKDOOR:
            m_obs.backdoor_tried = True
        return m_obs

    @staticmethod
    def install_tools_tried(a: EmulationAttackerAction, m_obs: EmulationAttackerMachineObservationState):
        """
        Utility function for updating install-tools flag

        :param a: the action
        :param m_obs: the observation to update
        :return: the updated observation
        """
        if a.id == EmulationAttackerActionId.INSTALL_TOOLS:
            m_obs.install_tools_tried = True
        return m_obs

    @staticmethod
    def cache_attacker_action(a: EmulationAttackerAction, s: EmulationEnvState) \
            -> None:
        """
        Utility function for caching an attacker action

        :param a: the attacker action to cache
        :param s: the current state
        :return: None
        """
        if s.attacker_obs_state is None:
            raise ValueError("EmulationAttackerObservationState is None")
        logged_in_ips_str = EnvDynamicsUtil.logged_in_ips_str(emulation_env_config=s.emulation_env_config, s=s)
        s.attacker_obs_state.actions_tried.add((a.id, a.index, logged_in_ips_str))

    @staticmethod
    def cache_defender_action(a: EmulationDefenderAction, s: EmulationEnvState) \
            -> None:
        """
        Utility function for caching a defender action

        :param a: the defender action to cache
        :param s: the current state
        :return: None
        """
        if s.defender_obs_state is None:
            raise ValueError("EmulationAttackerObservationState is None")
        logged_in_ips_str = EnvDynamicsUtil.logged_in_ips_str(emulation_env_config=s.emulation_env_config, s=s)
        s.defender_obs_state.actions_tried.add((int(a.id.value), a.index, logged_in_ips_str))

    @staticmethod
    def logged_in_ips_str(emulation_env_config: EmulationEnvConfig, s: EmulationEnvState) -> str:
        """
        Utility function to getting a string-id of the attacker state (Q) of logged in machines

        :param emulation_env_config: the environment config
        :param s: the current state
        :return: the string id
        """
        hacker_ip = emulation_env_config.containers_config.agent_ip
        if s.attacker_obs_state is None:
            raise ValueError("EmultaionAttackerObservationState is None")
        machines = s.attacker_obs_state.machines
        logged_in_ips = list(map(lambda x: "_".join(x.ips) + "_tools=" + str(int(x.tools_installed)) + "_backdoor="
                                           + str(int(x.backdoor_installed)) + "_root=" + str(int(x.root)),
                                 filter(lambda x: x.logged_in, machines)))
        logged_in_ips.append(hacker_ip)
        logged_in_ips = sorted(logged_in_ips, key=lambda x: x)
        logged_in_ips_str = "_".join(logged_in_ips)
        return logged_in_ips_str

    @staticmethod
    def exploit_get_vuln_name(a: EmulationAttackerAction) -> str:
        """
        Utiltiy function to get the vulnerability name of a particular exploit action

        :param a: the action
        :return: the name of hte corresponding vulnerability
        """
        if a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.FTP_DICT_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.SSH_DICT_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.TELNET_DICTS_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL:
            return constants.EXPLOIT_VULNERABILITES.IRC_DICTS_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.POSTGRES_DICTS_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.SMTP_DICTS_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.MYSQL_DICTS_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.MONGO_DICTS_SAME_USER_PASS
        elif a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.EXPLOIT_VULNERABILITES.CASSANDRA_DICTS_SAME_USER_PASS
        return constants.EXPLOIT_VULNERABILITES.UNKNOWN

    @staticmethod
    def exploit_get_vuln_cvss(a: EmulationAttackerAction) -> float:
        """
        Utility function for getting the CVSS of an exploit action

        :param a: the action
        :return: the CVSS
        """
        if a.id in EmulationAttackerActionConfig.dict_brute_same_user_ids():
            return constants.EXPLOIT_VULNERABILITES.WEAK_PASSWORD_CVSS
        return 0.0

    @staticmethod
    def exploit_get_service_name(a: EmulationAttackerAction) -> str:
        """
        Utility function to get the name of the exploited service of a particular exploit action

        :param a: the action
        :return: the name of the service
        """
        if a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.FTP.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.SSH.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.TELNET.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.IRC.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.POSTGRES.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.SMTP.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.MYSQL.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.MONGO.SERVICE_NAME
        elif a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL \
                or a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST:
            return constants.CASSANDRA.SERVICE_NAME
        return constants.EXPLOIT_VULNERABILITES.UNKNOWN

    @staticmethod
    def merge_duplicate_machines(machines: List[EmulationAttackerMachineObservationState],
                                 action: EmulationAttackerAction) -> List[EmulationAttackerMachineObservationState]:
        """
        Utility function for merging machines that are duplicates

        :param machines: list of machines (possibly with duplicates)
        :param action: the action that generated the new machines
        :return: the merged set of machines
        """
        merged_machines = []
        ips: Set[str] = set()
        for m in machines:
            ip_suffixes = list(map(lambda x: x.split(".")[-1], m.ips))
            for m1 in machines:
                for ip in m1.ips:
                    if ip.split(".")[-1] in ip_suffixes and ip not in m.ips:
                        m.ips = list(set(m.ips + m1.ips))
        for m in machines:
            if not m.ips_match(list(ips)):
                merged_m = m
                for m2 in machines:
                    if m2.ips == merged_m.ips:
                        merged_m = EnvDynamicsUtil.merge_new_machine_obs_with_old_machine_obs(merged_m, m2, action)
                merged_machines.append(merged_m)
                ips = ips.union(set(m.ips))
        return merged_machines



from typing import List, Dict, Any
from csle_common.dao.emulation_config.user import User
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeUsersConfig(JSONSerializable):
    """
    A DTO object representing the users of a container in an emulation environment
    """

    def __init__(self, ip: str, users: List[User], docker_gw_bridge_ip: str = "", physical_host_ip: str = ""):
        """
        Initializes the DTO

        :param ip: the ip of the node
        :param users: the list of users
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.ip = ip
        self.users = users
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeUsersConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NodeUsersConfig(
            ip=d["ip"], users=list(map(lambda x: User.from_dict(x), d["users"])),
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"],
            physical_host_ip=d["physical_host_ip"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["users"] = list(map(lambda x: x.to_dict(), self.users))
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, docker_gw_bridge_ip:{self.docker_gw_bridge_ip}, " \
               f"users:{','.join(list(map(lambda x: str(x), self.users)))}, physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeUsersConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeUsersConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeUsersConfig":
        """
        :return: a copy of the DTO
        """
        return NodeUsersConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeUsersConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        return config



from typing import Optional, List, Dict, Any
import json
import os
import csle_common.constants.constants as constants
from csle_common.dao.encoding.np_encoder import NpEncoder
from csle_base.json_serializable import JSONSerializable


class SimulationTrace(JSONSerializable):
    """
    Class that represents a trace in the simulation system
    """

    def __init__(self, simulation_env: str):
        """
        Initializes the DTO
        """
        self.simulation_env = simulation_env
        self.attacker_rewards: List[float] = []
        self.defender_rewards: List[float] = []
        self.attacker_observations: List[Any] = []
        self.defender_observations: List[Any] = []
        self.infos: List[Any] = []
        self.dones: List[Any] = []
        self.attacker_actions: List[Any] = []
        self.defender_actions: List[Any] = []
        self.states: List[Any] = []
        self.beliefs: List[Any] = []
        self.infrastructure_metrics: List[Any] = []
        self.id: int = -1

    def __str__(self) -> str:
        """
        :return: a string representation of the trace
        """
        return f"simulation_env: {self.simulation_env}, attacker_rewards:{self.attacker_rewards}, " \
               f"defender_rewards:{self.defender_rewards}, attacker_actions:{self.attacker_actions}, " \
               f"defender_actions:{self.defender_actions}, " \
               f"attacker_observations:{self.attacker_observations}, " \
               f"defender_observations:{self.defender_observations}, " \
               f"infos:{self.infos}, dones:{self.dones}, " \
               f"states: {self.states}, beliefs: {self.beliefs}, " \
               f"infrastructure_metrics: {self.infrastructure_metrics}, id: {self.id}"

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the trace
        """
        return {
            "simulation_env": self.simulation_env,
            "attacker_rewards": self.attacker_rewards,
            "defender_rewards": self.defender_rewards,
            "attacker_observations": self.attacker_observations,
            "defender_observations": self.defender_observations,
            "infos": self.infos,
            "dones": self.dones,
            "attacker_actions": self.attacker_actions,
            "defender_actions": self.defender_actions,
            "states": self.states,
            "beliefs": self.beliefs,
            "infrastructure_metrics": self.infrastructure_metrics,
            "id": self.id
        }

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SimulationTrace":
        """
        Converts a dict representation of the trace to a DTO representation

        :param d: the dict to convert
        :return: the trace DTO
        """
        trace = SimulationTrace(simulation_env=d["simulation_env"])
        if "attacker_rewards" in d:
            trace.attacker_rewards = d["attacker_rewards"]
        if "defender_rewards" in d:
            trace.defender_rewards = d["defender_rewards"]
        if "attacker_observations" in d:
            trace.attacker_observations = d["attacker_observations"]
        if "defender_observations" in d:
            trace.defender_observations = d["defender_observations"]
        if "infos" in d:
            trace.infos = d["infos"]
        if "dones" in d:
            trace.dones = d["dones"]
        if "attacker_actions" in d:
            trace.attacker_actions = d["attacker_actions"]
        if "defender_actions" in d:
            trace.defender_actions = d["defender_actions"]
        if "beliefs" in d:
            trace.beliefs = d["beliefs"]
        if "states" in d:
            trace.states = d["states"]
        if "infrastructure_metrics" in d:
            trace.infrastructure_metrics = d["infrastructure_metrics"]
        if "id" in d:
            trace.id = d["id"]
        return trace

    @staticmethod
    def save_traces(traces_save_dir, traces: List["SimulationTrace"], traces_file: Optional[str] = None) -> None:
        """
        Utility function for saving a list of traces to a json file

        :param traces_save_dir: the directory where to save the traces
        :param traces: the traces to save
        :param traces_file: the filename of the traces file
        :return: None
        """
        if traces_file is None:
            traces_file = constants.SYSTEM_IDENTIFICATION.SIMULATION_TRACES_FILE
        traces_dicts = list(map(lambda x: x.to_dict(), traces))
        if not os.path.exists(traces_save_dir):
            os.makedirs(traces_save_dir)
        with open(traces_save_dir + constants.COMMANDS.SLASH_DELIM + traces_file, 'w') as fp:
            json.dump({"traces": traces_dicts}, fp, cls=NpEncoder)

    @staticmethod
    def load_traces(traces_save_dir: str, traces_file: Optional[str] = None) -> List["SimulationTrace"]:
        """
        Utility function for loading and parsing a list of traces from a json file

        :param traces_save_dir: the directory where to load the traces from
        :param traces_file: (optional) a custom name of the traces file
        :return: a list of the loaded traces
        """
        if traces_file is None:
            traces_file = constants.SYSTEM_IDENTIFICATION.SIMULATION_TRACES_FILE
        path = traces_save_dir + constants.COMMANDS.SLASH_DELIM + traces_file
        if os.path.exists(path):
            with open(path, 'r') as fp:
                d = json.load(fp)
                traces: List[Dict[str, Any]] = d["traces"]
                return list(map(lambda x: SimulationTrace.from_dict(x), traces))
        else:
            print("Warning: Could not read traces file, path does not exist:{}".format(path))
            return []

    @staticmethod
    def from_json_file(json_file_path: str) -> "SimulationTrace":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SimulationTrace.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeBeatsConfig(JSONSerializable):
    """
    A DTO object representing the beats configuration of an individual container in an emulation
    """

    def __init__(self, ip: str, log_files_paths: List[str], filebeat_modules: List[str],
                 metricbeat_modules: List[str], heartbeat_hosts_to_monitor: List[str],
                 kafka_input: bool = False,
                 start_filebeat_automatically: bool = False, start_packetbeat_automatically: bool = False,
                 start_metricbeat_automatically: bool = False, start_heartbeat_automatically: bool = False):
        """
        Initializes the DTO

        :param ip: the ip of the nod
        :param log_files_paths: list of log files to ingest to elastic through filebeat
        :param filebeat_modules: list of filebeat modules to enable
        :param metricbeat_modules: list of metricbeat modules to enable
        :param heartbeat_hosts_to_monitor: list of hosts to monitor with heartbeat
        :param kafka_input: boolean indicating whether the kafka log should be ingested from this node or not
        :param start_filebeat_automatically: boolean indicating whether filebeat should be started automatically
                                             when the emulation is started
        :param start_packetbeat_automatically: boolean indicating whether packetbeat should be started automatically
                                               when the emulation is started
        :param start_metricbeat_automatically: boolean indicating whether metricbeat should be started automatically
                                               when the emulation is started
        :param start_heartbeat_automatically: boolean indicating whether heartbeat should be started automatically
                                               when the emulation is started
        """
        self.ip = ip
        self.log_files_paths = log_files_paths
        self.filebeat_modules = filebeat_modules
        self.kafka_input = kafka_input
        self.start_filebeat_automatically = start_filebeat_automatically
        self.start_packetbeat_automatically = start_packetbeat_automatically
        self.metricbeat_modules = metricbeat_modules
        self.start_metricbeat_automatically = start_metricbeat_automatically
        self.start_heartbeat_automatically = start_heartbeat_automatically
        self.heartbeat_hosts_to_monitor = heartbeat_hosts_to_monitor

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeBeatsConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NodeBeatsConfig(
            ip=d["ip"], log_files_paths=d["log_files_paths"], filebeat_modules=d["filebeat_modules"],
            kafka_input=d["kafka_input"], start_filebeat_automatically=d["start_filebeat_automatically"],
            start_packetbeat_automatically=d["start_packetbeat_automatically"],
            metricbeat_modules=d["metricbeat_modules"],
            start_metricbeat_automatically=d["start_metricbeat_automatically"],
            start_heartbeat_automatically=d["start_heartbeat_automatically"],
            heartbeat_hosts_to_monitor=d["heartbeat_hosts_to_monitor"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["log_files_paths"] = self.log_files_paths
        d["filebeat_modules"] = self.filebeat_modules
        d["kafka_input"] = self.kafka_input
        d["start_filebeat_automatically"] = self.start_filebeat_automatically
        d["start_packetbeat_automatically"] = self.start_packetbeat_automatically
        d["metricbeat_modules"] = self.metricbeat_modules
        d["start_metricbeat_automatically"] = self.start_metricbeat_automatically
        d["start_heartbeat_automatically"] = self.start_heartbeat_automatically
        d["heartbeat_hosts_to_monitor"] = self.heartbeat_hosts_to_monitor
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, log_files_paths: {self.log_files_paths}, filebeat_modules: {self.filebeat_modules}, " \
               f"kafka_input: {self.kafka_input}, start_filebeat_automatically: {self.start_filebeat_automatically}, " \
               f"start_packetbeat_automatically: {self.start_packetbeat_automatically}, " \
               f"metricbeat_modules: {self.metricbeat_modules}, " \
               f"start_metricbeat_automatically: {self.start_metricbeat_automatically}," \
               f"start_heartbeat_automatically: {self.start_heartbeat_automatically}," \
               f"heartbeat_hosts_to_monitor: {self.heartbeat_hosts_to_monitor}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeBeatsConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeBeatsConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeBeatsConfig":
        """
        :return: a copy of the DTO
        """
        return NodeBeatsConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeBeatsConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        heartbeat_hosts_to_monitor = []
        for host in self.heartbeat_hosts_to_monitor:
            heartbeat_hosts_to_monitor.append(GeneralUtil.replace_first_octet_of_ip(ip=host,
                                                                                    ip_first_octet=ip_first_octet))
        config.heartbeat_hosts_to_monitor = heartbeat_hosts_to_monitor
        return config



import os
from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.util.cluster_util import ClusterUtil


class TestClusterUtilSuite:
    """
    Test suite for cluster util
    """

    def test_am_i_leader(self) -> None:
        """
        Test function that checks if a given IP is a leader or not
        """
        config = MagicMock()
        config.cluster_config.cluster_nodes.ip = "192.168.1.1"
        config.cluster_config.cluster_nodes.leader = False
        config.cluster_config.cluster_nodes.cpus = 1
        config.cluster_config.cluster_nodes.gpus = 2
        config.cluster_config.cluster_nodes.RAM = 8
        assert not ClusterUtil.am_i_leader("192.168.1.1", config)

    @patch("psycopg.connect")
    @patch("csle_common.util.cluster_util.ClusterUtil.get_config")
    @patch.dict(os.environ, {"CSLE_HOME": "/mock/csle/home"})
    def test_get_config(self, mock_get_config, mock_connect) -> None:
        """
        Test function that gets the current cluster config from the metastore or from disk
        depending on if it is the leader node or not

        :param mock_get_config: mock_get_config
        :param mock_connect: mock_connect
        :return: None
        """
        mock_config = MagicMock()
        mock_get_config.return_value = mock_config
        from csle_common.util.cluster_util import ClusterUtil

        result = ClusterUtil.get_config()
        assert result == mock_config

    @patch.dict(os.environ, {"CSLE_HOME": "/mock/csle/home"})
    @patch("csle_common.util.cluster_util.ClusterUtil.get_config")
    def test_set_config_parameters_from_config_file(self, mock_get_config) -> None:
        """
        Test function that reads the config file from $CSLE_HOME/config.json and initializes certain config parameters

        :param mock_get_config: mock_get_config
        :return: None
        """
        mock_config = MagicMock()
        mock_config.management_admin_username_default = "admin"
        mock_config.management_admin_password_default = "admin_pw"
        mock_config.management_admin_email_default = "admin@example.com"
        mock_config.management_admin_organization_default = "Example Org"
        mock_config.management_admin_first_name_default = "Admin"
        mock_config.management_admin_last_name_default = "User"
        mock_config.ssh_admin_username = "ssh_admin"
        mock_config.ssh_admin_password = "ssh_pw"
        mock_config.management_guest_username_default = "guest"
        mock_config.management_guest_password_default = "guest_pw"
        mock_config.management_guest_email_default = "guest@example.com"
        mock_config.management_guest_organization_default = "Example Org"
        mock_config.management_guest_first_name_default = "Guest"
        mock_config.management_guest_last_name_default = "User"
        mock_config.ssh_agent_username = "agent"
        mock_config.ssh_agent_password = "agent_pw"
        mock_config.metastore_user = "metastore_user"
        mock_config.metastore_password = "metastore_pw"
        mock_config.metastore_database_name = "metastore_db"
        mock_config.pgadmin_username = "pgadmin"
        mock_config.pgadmin_password = "pgadmin_pw"
        mock_config.grafana_username = "grafana"
        mock_config.grafana_password = "grafana_pw"
        mock_config.node_exporter_port = 9100
        mock_config.grafana_port = 3000
        mock_config.management_system_port = 8000
        mock_config.cadvisor_port = 8080
        mock_config.pgadmin_port = 5050
        mock_config.prometheus_port = 9090
        mock_config.node_exporter_pid_file = "/var/run/node_exporter.pid"
        mock_config.csle_mgmt_webapp_pid_file = "/var/run/csle_mgmt_webapp.pid"
        mock_config.node_exporter_log_file = "/var/log/node_exporter.log"
        mock_config.docker_stats_manager_outfile = "/var/log/docker_stats.out"
        mock_config.docker_stats_manager_pidfile = "/var/run/docker_stats.pid"
        mock_config.prometheus_pid_file = "/var/run/prometheus.pid"
        mock_config.prometheus_log_file = "/var/log/prometheus.log"
        mock_config.postgresql_log_dir = "/var/log/postgresql"
        mock_config.nginx_log_dir = "/var/log/nginx"
        mock_config.flask_log_file = "/var/log/flask.log"
        mock_config.default_log_dir = "/var/log"
        mock_get_config.return_value = mock_config
        ClusterUtil.set_config_parameters_from_config_file()
        assert constants.CONFIG_FILE.PARSED_CONFIG == mock_config
        assert constants.CSLE_ADMIN.MANAGEMENT_USER == "admin"
        assert constants.CSLE_ADMIN.MANAGEMENT_PW == "admin_pw"
        assert constants.CSLE_ADMIN.MANAGEMENT_EMAIL == "admin@example.com"
        assert constants.CSLE_ADMIN.MANAGEMENT_ORGANIZATION == "Example Org"
        assert constants.CSLE_ADMIN.MANAGEMENT_FIRST_NAME == "Admin"
        assert constants.CSLE_ADMIN.MANAGEMENT_LAST_NAME == "User"
        assert constants.CSLE_ADMIN.SSH_USER == "ssh_admin"
        assert constants.CSLE_ADMIN.SSH_PW == "ssh_pw"
        assert constants.CSLE_GUEST.MANAGEMENT_USER == "guest"
        assert constants.CSLE_GUEST.MANAGEMENT_PW == "guest_pw"
        assert constants.CSLE_GUEST.MANAGEMENT_EMAIL == "guest@example.com"
        assert constants.CSLE_GUEST.MANAGEMENT_ORGANIZATION == "Example Org"
        assert constants.CSLE_GUEST.MANAGEMENT_FIRST_NAME == "Guest"
        assert constants.CSLE_GUEST.MANAGEMENT_LAST_NAME == "User"
        assert constants.AGENT.USER == "agent"
        assert constants.AGENT.PW == "agent_pw"
        assert constants.METADATA_STORE.USER == "metastore_user"
        assert constants.METADATA_STORE.PASSWORD == "metastore_pw"
        assert constants.METADATA_STORE.DBNAME == "metastore_db"
        assert constants.COMMANDS.PGADMIN_USERNAME == "pgadmin"
        assert constants.COMMANDS.PGADMIN_PW == "pgadmin_pw"
        assert constants.COMMANDS.GRAFANA_USERNAME == "grafana"
        assert constants.COMMANDS.GRAFANA_PW == "grafana_pw"
        assert constants.COMMANDS.NODE_EXPORTER_PORT == 9100
        assert constants.COMMANDS.GRAFANA_PORT == 3000
        assert constants.COMMANDS.MANAGEMENT_SYSTEM_PORT == 8000
        assert constants.COMMANDS.CADVISOR_PORT == 8080
        assert constants.COMMANDS.PGADMIN_PORT == 5050
        assert constants.COMMANDS.PROMETHEUS_PORT == 9090
        assert constants.COMMANDS.NODE_EXPORTER_PID_FILE == "/var/run/node_exporter.pid"
        assert constants.COMMANDS.CSLE_MGMT_WEBAPP_PID_FILE == "/var/run/csle_mgmt_webapp.pid"
        assert constants.COMMANDS.NODE_EXPORTER_LOG_FILE == "/var/log/node_exporter.log"
        assert constants.COMMANDS.DOCKER_STATS_MANAGER_OUTFILE == "/var/log/docker_stats.out"
        assert constants.COMMANDS.DOCKER_STATS_MANAGER_PIDFILE == "/var/run/docker_stats.pid"
        assert constants.COMMANDS.PROMETHEUS_PID_FILE == "/var/run/prometheus.pid"
        assert constants.COMMANDS.PROMETHEUS_LOG_FILE == "/var/log/prometheus.log"
        assert constants.COMMANDS.POSTGRESQL_LOG_DIR == "/var/log/postgresql"
        assert constants.COMMANDS.NGINX_LOG_DIR == "/var/log/nginx"
        assert constants.COMMANDS.FLASK_LOG_FILE == "/var/log/flask.log"
        assert constants.LOGGING.DEFAULT_LOG_DIR == "/var/log"



from typing import Optional, List, Dict, Any
import random
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential
from csle_base.json_serializable import JSONSerializable


class NetworkService(JSONSerializable):
    """
    DTO Class representing a service in the network
    """

    def __init__(self, protocol: TransportProtocol, port: int, name: str,
                 credentials: Optional[List[Credential]] = None):
        """
        Initializes the DTO

        :param protocol: the protocol of the service
        :param port: the port of the service
        :param name: the name of the service
        :param credentials: the list of credentials of the service
        """
        self.protocol = protocol
        self.port = port
        self.name = name
        self.credentials = credentials

    def __str__(self) -> str:
        """
        :return: a string representation of the service
        """
        cr = []
        if self.credentials is not None:
            cr = list(map(lambda x: str(x), self.credentials))
        return f"protocol: {self.protocol}, port: {self.port}, name: {self.name}, cr: {cr}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["protocol"] = self.protocol
        d["port"] = self.port
        d["name"] = self.name
        if self.credentials is not None:
            d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials))
        else:
            d["credentials"] = self.credentials
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NetworkService":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        credentials = None
        if d["credentials"] is not None:
            credentials = list(map(lambda x: Credential.from_dict(x), d["credentials"]))
        dto = NetworkService(name=d["name"], port=d["port"], protocol=d["protocol"], credentials=credentials)
        return dto

    def copy(self) -> "NetworkService":
        """
        :return: a copy of the DTO
        """
        return NetworkService(protocol=self.protocol, port=self.port, name=self.name, credentials=self.credentials)

    @staticmethod
    def pw_vuln_services():
        """
        :return: a list of all vulnerabilities that involve weak passwords
        """
        ssh_vuln_service = (NetworkService(protocol=TransportProtocol.TCP, port=22, name="ssh", credentials=[]),
                            constants.EXPLOIT_VULNERABILITES.SSH_DICT_SAME_USER_PASS)
        ftp_vuln_service = (NetworkService(protocol=TransportProtocol.TCP, port=21, name="ftp", credentials=[]),
                            constants.EXPLOIT_VULNERABILITES.FTP_DICT_SAME_USER_PASS)
        telnet_vuln_service = (NetworkService(protocol=TransportProtocol.TCP, port=23, name="telnet", credentials=[]),
                               constants.EXPLOIT_VULNERABILITES.TELNET_DICTS_SAME_USER_PASS)
        return [ssh_vuln_service, ftp_vuln_service, telnet_vuln_service], [ssh_vuln_service, telnet_vuln_service]

    def __eq__(self, other) -> bool:
        """
        Tests equality with another service

        :param other: the service to compare with
        :return: True if equal otherwise False
        """
        if not isinstance(other, NetworkService):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.name == other.name and self.protocol == other.protocol and self.port == other.port

    @staticmethod
    def from_credential(credential: Credential) -> "NetworkService":
        """
        Converts the object into a network service representation

        :return: the network service representation
        """
        if credential.protocol is None:
            protocol = TransportProtocol(random.choice([0, 1]))
        else:
            protocol = credential.protocol
        if credential.service is None:
            name = ""
        else:
            name = credential.service
        service = NetworkService(protocol=protocol, port=credential.port,
                                 name=name, credentials=[credential])
        return service

    @staticmethod
    def from_json_file(json_file_path: str) -> "NetworkService":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NetworkService.from_dict(json.loads(json_str))



from enum import IntEnum


class EmulationDefenderActionId(IntEnum):
    """
    Enum representing the different defenses
    """
    STOP = 0
    CONTINUE = 1
    RESET_USERS = 2
    ENABLE_DPI = 3
    BLACKLIST_IPS = 4
